EXECUTE = "Применить"
REPLY_TO_HLS_AUTHENTICATION = "Ответ на проверку подлинности"
CHANGE_HLS_SECRET = "Изменение пароля"
ADD_OBJECT = "Добавление объекта"
REMOVE_OBJECT = "Удаление объекта"
ACTIVATE_PASSIVE_CALENDAR = "Активировать пассивный календарь"
ADJUST_TO_QUARTER = "Подстройка к ¼ часа"
ADJUST_TO_MEASURING_PERIOD = "Подстройка к измерительному периоду"
ADJUST_TO_MINUTE = "Подстройка к минуте"
ADJUST_TO_PRESET_TIME = "Подстройка к уставке времени"
PRESET_ADJUSTING_TIME = "Задание уставки времени"
SHIFT_TIME = "Сдвиг времени"
REMOTE_DISCONNECT = "Отключение"  # in origin "Удаленное отключение"
REMOTE_RECONNECT = "Подключение"  # in origin "Удаленное включение"
RESET = "Сброс"
IMAGE_TRANSFER_INITIATE = "Инициализация передачи"
IMAGE_BLOCK_TRANSFER = "Передача блока данных"
IMAGE_VERIFY = "Проверка файла"
IMAGE_ACTIVATE = "Активирование файла"
CAPTURE = "Захват"
ENABLE_DISABLE = "Разрешить/запретить (номера)"
INSERT = "Ввести (номер)"
DELETE = "Стереть (номер)"
SECURITY_ACTIVATE = "Усиление безопасности"
GLOBAL_KEY_TRANSFER = "Передача общего ключа"
KEY_TRANSFER = "Передача ключа"
KEY_AGREEMENT = "Ключевые соглашения"
GENERATE_KEY_PAIR = "Генерация пары для ключа"
GENERATE_CERTIFICATE_REQUEST = "Запрос генерации сертификата"
IMPORT_CERTIFICATE = "Импорт сертификата"
EXPORT_CERTIFICATE = "Экспорт сертификата"
REMOVE_CERTIFICATE = "Удаление сертификата"
REQUEST_ACTION = "Запрос действия"
PUSH = "Протолнуть"
# demand register
NEXT_PERIOD = "Следующий период"
