ASSOCIATION_LN = "Association LN"
ARBITRATOR = "Arbitrator"
PSTN_MODEM_CONFIGURATION = "PSTN modem configuration"
MODEM_CONFIGURATION = "Modem configuration"
SECURITY_SETUP = "Security setup"
ACTIVITY_CALENDAR = "Activity calendar"
CLIENT_SETUP = "Client setup"
CLOCK = "Clock"
DATA = "Data"
DISCONNECT_CONTROL = "Disconnect control"
EXTENDED_REGISTER = "Extended register"
DEMAND_REGISTER = "Demand register"
GPRS_MODEM_SETUP = "GPRS modem setup"
GSM_DIAGNOSTIC = "GSM diagnostic"
IEC_HDLC_SETUP = "IEC HDLC setup"
IEC_LOCAL_PORT_SETUP = "IEC local port setup"
IMAGE_TRANSFER = "Image transfer"
IPV4_SETUP = "IPv4 setup"
LIMITER = "Limiter"
PROFILE_GENERIC = "Profile generic"
PUSH_SETUP = "Push setup"
REGISTER = "Register"
REGISTER_MONITOR = "Register monitor"
SCHEDULE = "Schedule"
SCRIPT_TABLE = "Script table"
SINGLE_ACTION_SCHEDULE = "Single action schedule"
SPECIAL_DAYS_TABLE = "Special days table"
TCP_UDP_SETUP = "TCP UDP setup"
